package lu.uni.timed_mutation.refinementparsing;


import java.io.IOException;
import java.nio.file.Paths;
import java.util.Scanner;

/**
 * 
 * @author Davide Basile
 * @author Maxime Cordy
 *
 */
public class AppEcdar {

	public enum MutantValue{
		NONSUBSUMED, SUBSUMED, INCONSISTENT, NIL
	}	

	private static String casestudy="COFFEE";
	private static boolean refinement = false;
	private static boolean parselogs = false;
	private static int order = 1;
	public static String pathSeparator = System.getProperty("os.name").toLowerCase().contains("win")?"\\":"/";
	private static String ecdarpath=".."+pathSeparator;
	private static String mutantpath=".."+pathSeparator;
	

	public static void main(String[] args) throws IOException {
		//				casestudy="COFFEE";
		//				mutantpath+=casestudy+"2\\";
		//				runRefinementChecking(2);
		//				parseLog(2,false,parseViolationsLog(2));

		
		if (args==null || args.length==0) {
			System.out.println("Usage : java -jar AppEcdar.jar [options] "+System.lineSeparator()+""
					+ "where options are :");
			System.out.println("-ecdarpath [path] to indicate the path where verifytga.exe is located, default is "+ecdarpath);
			System.out.println("-mutantpath [path] to indicate the path where the directories of the mutants generated by AppMutant are, default is "+mutantpath);
			System.out.println("-casestudy [COFFEE|CAS|WFAS|ACCEL|MUTEX|HOTEL] to select one of the available case studies, default is: "+casestudy);
			System.out.println("-refinement to perform  mutants refinement checking");
			System.out.println("-parselogs to parse logs of mutants and violations");
			System.out.println("-secondorder operations performed on second order mutants, default is first order");
			return;
		}
		
		for(int i = 0; i < args.length; i++) {
			if(args[i].equals("-ecdarpath")) {
				ecdarpath = args[++i];
			}
			else if (args[i].equals("-mutantpath")) {
				mutantpath= args[++i];
			}

			else if (args[i].equals("-casestudy")) {
				casestudy = args[++i];
				if (!(casestudy.equals("COFFEE")||casestudy.equals("CAS")||casestudy.equals("WFAS")
						||casestudy.equals("ACCEL")||casestudy.equals("MUTEX")||casestudy.equals("HOTEL")))
				{
					System.out.println("Case study not available");
					return;
				}
				
			}

			else if (args[i].equals("-refinement")) {
				refinement=true;
				
			}

			else if (args[i].equals("-parselogs")) {
				parselogs=true;
			}
			else if (args[i].equals("-secondorder"))
				order = 2;
		}


		mutantpath+=casestudy+order+pathSeparator;
		
		System.out.println("The folder where mutants and violations log generated by AppMutant are searched is "
				+ Paths.get(mutantpath).toAbsolutePath());
		
		if (refinement) {
			System.out.println(((order==1)?"First":"Second")+" order mutants refinement checking. "+System.lineSeparator()+" "
					+ "Warning! This operation may take a while. "+System.lineSeparator()+" "
					+ "Press any key to continue");
			Scanner scan = new Scanner(System.in); 
			scan.nextLine();
			scan.close();
			RefinementChecker rc = new RefinementChecker(mutantpath,ecdarpath);
			rc.accept(order);
			
		}
		
		if (parselogs) {
			System.out.println(((order==1)?"First":"Second")+" order  parsing refinement and violations logs. "+System.lineSeparator());
			LogParser lg = new LogParser(mutantpath,casestudy);
			lg.accept(order);
					
		}
		System.out.println("Terminated successfully");
	}

}







//System.out.println("-f  run Refinement Checking on first order mutants and  parse logs");
//System.out.println("-s  run Refinement Checking on second order mutants and parse logs");
//System.out.println("-pi  parse logs of first order mutants");
//System.out.println("-pe  parse logs of second order mutants");
//System.out.println("-a   run everything for all case studies");

//Scanner scan = new Scanner(System.in); 
//if (args[0].contains("a")) {
//	System.out.println("Everything. "+System.lineSeparator()+""
//			+ "Warning! This operation may take a while. "+System.lineSeparator()+" "
//			+ "Press any key to continue");
//	scan.nextLine();
//	System.out.println("WARNING, CONFIRM AGAIN! "+System.lineSeparator()+" ");
//	scan.nextLine();
//	String mutantpathback=mutantpath;
//	for (int i=3; i<7;i++)
//	{
//		String press = ""+i;
//		casestudy=(press.contains("1"))?"COFFEE"
//				:(press.contains("2"))?"CAS"
//						:(press.contains("3"))?"WFAS"
//								:(press.contains("4"))?"ACCEL"
//										:(press.contains("5"))?"MUTEX"
//												:(press.contains("6"))?"HOTEL"
//														:"";
//
//		System.out.println(casestudy+" selected.");
//
//		if (casestudy=="") {scan.close(); return;}
//		mutantpath+=casestudy;
//		mutantpath+="1\\";			
//		runRefinementChecking(1);
//		parseLog(1,parseViolationsLog(1));
//
//		mutantpath=mutantpathback;
//		mutantpath+=casestudy+"2\\";
//		runRefinementChecking(2);
//		parseLog(2,parseViolationsLog(2));
//
//		mutantpath=mutantpathback;
//	}
//	scan.close();
//	return;
//}
//
//System.out.println("Select Case study: "+System.lineSeparator()+""
//		+ "Type 1 for COFFEE "+System.lineSeparator()+""
//		+ "Type 2 for CAS "+System.lineSeparator()+""
//		+ "Type 3 for WFAS "+System.lineSeparator()+"" 
//		+ "Type 4 for ACCEL "+System.lineSeparator()+""
//		+ "Type 5 for MUTEX "+System.lineSeparator()+""
//		+ "Type 6 for HOTEL "+System.lineSeparator()+"");
//String press = scan.nextLine();
//casestudy=(press.contains("1"))?"COFFEE"
//		:(press.contains("2"))?"CAS"
//				:(press.contains("3"))?"WFAS"
//						:(press.contains("4"))?"ACCEL"
//								:(press.contains("5"))?"MUTEX"
//										:(press.contains("6"))?"HOTEL"
//												:"";
//
//System.out.println(casestudy+" selected.");
//if (casestudy=="") {scan.close(); return;}
//mutantpath+=casestudy;
//
//if (args[0].contains("f")) {
//	System.out.println("First order mutants checking and parsing. "+System.lineSeparator()+" "
//			+ "Warning! This operation may take a while. "+System.lineSeparator()+" "
//			+ "Press any key to continue");
//	scan.nextLine();
//	mutantpath+="1\\";
//	runRefinementChecking(1);
//	parseLog(1,parseViolationsLog(1));
//}
//else if (args[0].contains("s")) {
//	System.out.println("Second order mutants checking and parsing. "+System.lineSeparator()+" "
//			+ "Warning! This operation may take a while. "+System.lineSeparator()+""
//			+ "Press any key to continue");
//	scan.nextLine();
//	mutantpath+="2\\";
//	runRefinementChecking(2);
//	parseLog(2,parseViolationsLog(2));
//}
//else if (args[0].contains("pi")){
//	mutantpath+="1\\";
//	parseLog(1,parseViolationsLog(1));
//}
//else if (args[0].contains("pe")){
//	mutantpath+="2\\";
//	parseLog(2,parseViolationsLog(2));
//}
//
//scan.close();